// created: 5.8.01 modified: 23.8.01 final clean-up: 24.4.02
// Stephan Dale
// data manipulation program

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "maverik.h"

/**************************************************************************************************/

/********************************/
/* internal function prototypes */
/********************************/

int loadFile(char *filename, int *type);
int numLines(char *filename);
int normalise(char *inFileName, int type, float shoulderWidth, float upperArmLength, float lowerArmLength);
int split(char *inFileName, float percentage);
int join(char *inFileName, char *filename);
int randomise(char *inFileName);
int removeHandRPY(char *inFileName, int type);
int clean(char *inFileName, int type);
int relBack(char *inFileName, int type, float shoulderWidth, float upperArmLength);
int calcError(char *inFileName, int type);

MAV_vector getElbowPoint(float yaw, float pitch, float shoulderWidth, float upperArmLength, float lowerArmLength);
MAV_vector IKAlgorithm(MAV_vector prediction, MAV_vector handPosition, float shoulderWidth, float upperArmLength, float lowerArmLength);
int angularToPositional(char *netoutFilename, int patternNumbers, char *testingFilename, float shoulderWidth, float upperArmLength, float lowerArmLength, int ikAlgorithm);

/*int tmpCorrect(char *inFileName, int type, char *filename);*/

/**************************************************************************************************/

int verbose = 0;

/**************************************************************************************************/

/******************/
/* load file name */
/******************/
// BUG WARNING: not really a bug, but there is no error checking here. The info provided by the user is assumed to be true.
int loadFile(char *inFileName, int *type)
 {
 FILE *datain;
 char choice;

 printf("training or (testing) results data? (t or e): ");
 scanf("%s", &choice);
 if (choice=='t')
   {
   printf("does data contain hand RPY? (y or n): ");
   scanf("%s", &choice);
   if (choice=='y')
     *type = 1;
   else if (choice=='n')
     *type = 0;
   else
     {
     printf("not loading:- must specify whether data contains hand RPY\n");
     inFileName = NULL;
     *type = -1;
     return 0;
     }
   }
 else if (choice=='e')
   {
   printf("does data contain pattern number? (y or n): ");
   scanf("%s", &choice);
   if (choice=='n')
     *type = 2;
   else if (choice=='y')
     *type = 3;
   else
     {
     printf("not loading:- must specify whether data contains pattern number\n");
     inFileName = NULL;
     *type = -1;
     return 0;
     }
   }
 else
   {
   printf("not loading:- must specify whether training or (testing) results data\n");
   inFileName = NULL;
   *type = -1;
   return 0;
   }
 printf("please enter file name: ");
 scanf("%s", inFileName);
 if ((datain = fopen(inFileName,"r"))!=NULL)
   {
   printf("loading '%s'\n", inFileName);
   return 1;
   }
 else
   {
   printf("failed to load '%s'\n", inFileName);
   inFileName = NULL;
   return 0;
   }
 return 1;
 }

/*******************************/
/* get number of lines in file */
/*******************************/
int numLines(char *filename)
 {
 FILE *data;
 int count = 0;
 float tmp;
 char line[200]; // line can be up to 200 characters long

 if ((data = fopen(filename,"r"))!=NULL)
   {
   while (fgets(line,200,data) != NULL) // read line
     count++;
   }

 fclose(data);
 return count;
 }

/*************/
/* normalise */
/*************/
int normalise(char *inFileName, int type, float shoulderWidth, float upperArmLength, float lowerArmLength)
 {
 FILE *indata, *outdata;
 char str[200];
 MAV_vector vec, v, normal;
 float cosAngle, tmp, maxMag;
 float x, y, z, handRoll, handPitch, handYaw, armPitch, armYaw;

 if (type==2 || type==3)
  {
  printf("not normalising:- must load training data\n");
  return 0;
  }

 // try to create normalised data file
 strcpy(str, inFileName);
 strcat(str, ".norm");
 if ((outdata = fopen(str,"w"))!=NULL)
   {
   printf("writing normalised data to '%s'\n", str);
   }
 else
   {
   printf("failed to open '%s' for writing\n", str);
   return 0;
   }

 // normalise data using max arm length
 if ((indata = fopen(inFileName,"r"))!=NULL)
   {
   //  calculate max magnitude (using max possible arm extension)
   maxMag = shoulderWidth + upperArmLength + lowerArmLength;
   printf("max hand vector magnitude = %f\n", maxMag);
   // normalise hand vectors using max magnitude, and angles using max possible values (pitch 180, yaw 180)
   printf("normalising '%s'...", inFileName);
   while (!feof(indata))
     {
     if (type==1)
       {
       fscanf(indata, "%f %f %f %f %f %f %f %f", &x, &y, &z, &handRoll, &handPitch, &handYaw, &armYaw, &armPitch);
       // normalise hand RPY
       if (handRoll<0) handRoll += 360; // ensure >=0
       if (handPitch<0) handPitch += 360;
       if (handYaw<0) handYaw += 360;
       handRoll = handRoll/360.0; // max roll is 360
       handPitch = handPitch/360.0;
       handYaw = handYaw/360.0;
       }
     else
       fscanf(indata, "%f %f %f %f %f", &x, &y, &z, &armYaw, &armPitch);

     // normalise angles
     if (armYaw<0) armYaw += 360;
     if (armPitch<0) armPitch += 360; // ensure >=0
     armYaw = armYaw/360.0; // max yaw is 360
     armPitch = armPitch/360.0; // max pitch is 360

     // normalise hand vector
     vec = mav_vectorSet(x, y, z);
     tmp = mav_vectorMag(vec);
     vec = mav_vectorScalar(vec, (tmp/maxMag)/tmp);
     // map onto the range 0 -> 1 (from range -1 -> 1)
     // note: must remove any out of range (from innaccurate user proportions)?
     vec.x = (vec.x/2.0)+0.5;
     vec.y = (vec.y/2.0)+0.5;
     vec.z = (vec.z/2.0)+0.5;

     // write processed data
     if (type==1)
       fprintf(outdata, "%f %f %f %f %f %f %f %f\n", vec.x, vec.y, vec.z, handRoll, handPitch, handYaw, armYaw, armPitch);
     else
       fprintf(outdata, "%f %f %f %f %f\n", vec.x, vec.y, vec.z, armYaw, armPitch);
     }
   printf("done\n");
   fclose(indata);
   }
 else
   {
   printf("failed to open '%s'\n", inFileName);
   fclose(outdata);
   return 0;
   }

 fclose(outdata);
 return 1;
 }

/***********************/
/* split into two file */
/***********************/
int split(char *inFileName, float percentage)
 {
 FILE *indata, *outdata1, *outdata2;
 char str[200];
 float tmp;
 int count;

 // try to create data files for each part
 strcpy(str, inFileName);
 strcat(str, ".split1");
 if ((outdata1 = fopen(str,"w"))!=NULL)
   {
   printf("writing 1st part of data (%f%) to '%s'\n", percentage, str);
   }
 else
   {
   printf("failed to open '%s' for writing\n", str);
   return 0;
   }

 strcpy(str, inFileName);
 strcat(str, ".split2");
 if ((outdata2 = fopen(str,"w"))!=NULL)
   {
   printf("writing 2nd part of data (%f%) to '%s'\n", 100-percentage, str);
   }
 else
   {
   printf("failed to open '%s' for writing\n", str);
   fclose(outdata1);
   return 0;
   }

 tmp = (numLines(inFileName)/100.0)*percentage; // work out how often a line should be written to .split2 rather than .split1

 // split data
 printf("splitting '%s'...", inFileName);
 count = 0;
 if ((indata = fopen(inFileName,"r"))!=NULL)
   {

   while (fgets(str,200,indata) != NULL) // read line
     {
     count++;
     if (count<=tmp) // write line to either outdata1 or outdata2
       fputs(str, outdata1);
     else
       fputs(str, outdata2);
     }

   printf("done\n");
   fclose(indata);
   }
 else
   {
   printf("failed to open '%s'\n", inFileName);
   fclose(outdata1);
   fclose(outdata2);
   return 0;
   }

 fclose(outdata1);
 fclose(outdata2);
 return 1;
 }

/**************************/
/* join with another file */
/**************************/
// BUG WARNING: The first line of the second file (that you're joining to) is concatenated to the end of the last line of the first file.
//              This is not a problem as long as the last line of the first file is blank.
int join(char *inFileName, char *filename)
 {
 FILE *indata1, *indata2, *outdata;
 char str[200];
 float tmp;
 int count;

 // try to create data file
 strcpy(str, inFileName);
 strcat(str, ".join");
 if ((outdata = fopen(str,"w"))!=NULL)
   {
   printf("writing concatenated data to '%s'\n", str);
   }
 else
   {
   printf("failed to open '%s' for writing\n", str);
   return 0;
   }

 // join data
 printf("joining '%s' and '%s'...", inFileName, filename);
 count = 0;
 if ((indata1 = fopen(inFileName,"r"))!=NULL)
   {
   if ((indata2 = fopen(filename,"r"))!=NULL)
     {

     while (fgets(str,200,indata1) != NULL) // copy first file
       fputs(str, outdata);
     while (fgets(str,200,indata2) != NULL) // copy second file
       fputs(str, outdata);

     printf("done\n");
     fclose(indata2);
     }
   else
     printf("failed to open '%s'\n", filename);
   fclose(indata1);
   }
 else
   {
   printf("failed to open '%s'\n", inFileName);
   fclose(outdata);
   return 0;
   }

 fclose(outdata);
 return 1;
 }

/*******************/
/* randomise lines */
/*******************/
int randomise(char *inFileName)
 {
 FILE *indata, *outdata;
 char str[200], line[200];
 int i, j, count, maxLines;
 char **lines;
 char *tmp;
 float x, y, z, handRoll, handPitch, handYaw, armPitch, armYaw;

 // try to create randomised data file
 strcpy(str, inFileName);
 strcat(str, ".rand");
 if ((outdata = fopen(str,"w"))!=NULL)
   {
   printf("writing randomised data to '%s'\n", str);
   }
 else
   {
   printf("failed to open '%s' for writing\n", str);
   return 0;
   }

 // random number seed
 srandom(rand());

 // get number of lines in file and allocate array of lines
 maxLines = numLines(inFileName);
 lines = (char**) malloc(maxLines*sizeof(char*));
 count = 0;

 // randomise data using max arm length
 if ((indata = fopen(inFileName,"r"))!=NULL)
   {
   printf("randomising '%s'...",inFileName);

   // copy lines into array
   while (fgets(line,200,indata) != NULL)
     {
     lines[count] = (char*) malloc(strlen(line)+1);
     strcpy(lines[count],line);
     count++;
     }

   // shuffle the lines	
   for (i=0; i<maxLines; i++)
     {
     j = random() % maxLines;
     tmp = lines[j];
     lines[j] = lines[i];
     lines[i] = tmp;
     }

   // write shuffled lines
   for (i=0; i<maxLines; i++)
     fputs(lines[i], outdata);

   printf("done\n");
   fclose(indata);
   }
 else
   {
   printf("failed to open '%s'\n", inFileName);
   fclose(outdata);
   return 0;
   }

 fclose(outdata);
 return 1;
 }

/**********************************/
/* remove hand angular data (RPY) */
/**********************************/
int removeHandRPY(char *inFileName, int type)
 {
 FILE *indata, *outdata;
 char str[200];
 float x, y, z, handRoll, handPitch, handYaw, armPitch, armYaw;

 if (type==2 || type==3)
   {
   printf("not removing hand RPY:- must load training data\n");
   return 0;
   }

 if (type==0) // check that file contains hand RPY values
   {
   printf("not removing hand RPY:- file does not contain hand RPY\n");
   return 0;
   }

 // try to create data file
 strcpy(str, inFileName);
 strcat(str, ".noang");
 if ((outdata = fopen(str,"w"))!=NULL)
   {
   printf("writing all data but hand RPY to '%s'\n", str);
   }
 else
   {
   printf("failed to open '%s' for writing\n", str);
   return 0;
   }

 // remove hand RPY
 if ((indata = fopen(inFileName,"r"))!=NULL)
   {
   printf("removing hand RPY from '%s'...", inFileName);
   while (!feof(indata))
     {
     // read pattern
     fscanf(indata, "%f %f %f %f %f %f %f %f", &x, &y, &z, &handRoll, &handPitch, &handYaw, &armYaw, &armPitch);
     // write all but hand RPY
     fprintf(outdata, "%f %f %f %f %f\n", x, y, z, armYaw, armPitch);
     }
   printf("done\n");
   fclose(indata);
   }
 else
   {
   printf("failed to open '%s'\n", inFileName);
   fclose(outdata);
   return 0;
   }

 fclose(outdata);
 return 1;
 }

/**************/
/* clean data */
/**************/
int clean(char *inFileName, int type)
 {
 FILE *indata, *outdata;
 char str[200], previous[200];
 float x, y, z, handRoll, handPitch, handYaw, armPitch, armYaw;
// float previousSum;
 int count;

 if (type==2 || type==3)
   {
   printf("not cleaning:- must load training data\n");
   return 0;
   }

 // try to create data file
 strcpy(str, inFileName);
 strcat(str, ".clean");
 if ((outdata = fopen(str,"w"))!=NULL)
   printf("writing cleaned data to '%s'\n", str);
 else
   {
   printf("failed to open '%s' for writing\n", str);
   return 0;
   }

 // clean (remove duplicates)
 if ((indata = fopen(inFileName,"r"))!=NULL)
   {
   printf("cleaning %s...", inFileName);
   count = 0;
   while (fgets(str,200,indata) != NULL) // get line
     {
     count++;
     if (strcmp(str, previous)) // if not the same...
       {
       fputs(str, outdata); // ...keep
       }
     strcpy(previous, str);
     }

/*   previousSum = 1000; // set previous sum to impossible value (max possible = 360*2)
   while (!feof(indata))
     {
     if (type==1)
       {
       // read pattern
       fscanf(indata, "%f %f %f %f %f %f %f %f", &x, &y, &z, &handRoll, &handPitch, &handYaw, &armPitch, &armYaw);
       if ((armPitch+armYaw)!=previousSum) // if arm pitch and yaw have changed, write pattern to file
         fprintf(outdata, "%f %f %f %f %f %f %f %f\n", x, y, z, handRoll, handPitch, handYaw, armPitch, armYaw);
       }
     else
       {
       // read pattern
       fscanf(indata, "%f %f %f %f %f", &x, &y, &z, &armPitch, &armYaw);
       if ((armPitch+armYaw)!=previousSum) // if arm pitch and yaw have changed, write pattern to file
         fprintf(outdata, "%f %f %f %f %f\n", x, y, z, armPitch, armYaw);
       }
     previousSum = armPitch+armYaw;
     }*/

   printf("done\n");
   fclose(indata);
   }
 else
   {
   printf("failed to open '%s'\n", inFileName);
   fclose(outdata);
   return 0;
   }

 fclose(outdata);
 return 1;
 }

/*******************************************/
/* change arm rotation so relative to back */
/*******************************************/
int relBack(char *inFileName, int type, float shoulderWidth, float upperArmLength)
 {
 MAV_vector vec, v, normal;
 float cosAngle;
 FILE *indata, *outdata;
 char choice;
 char str[200], previous[200];
 float x, y, z, handRoll, handPitch, handYaw, armPitch, armYaw;
// float previousSum;
 int count;

 if (type==2)
   {
   printf("not changing arm rotation:- must load training data\n");
   return 0;
   }

 // try to create data file
 strcpy(str, inFileName);
 strcat(str, ".back");
 if ((outdata = fopen(str,"w"))!=NULL)
   printf("writing new arm rotation data to '%s'\n", str);
 else
   {
   printf("failed to open '%s' for writing\n", str);
   return 0;
   }

 // pre-process (get upper arm rotation relative to back position and rotation, with yaw relative to x-axis)
 if ((indata = fopen(inFileName,"r"))!=NULL)
   {
   printf("processing %s...", inFileName);
   while (!feof(indata))
     {
     // read pattern
     if (type==1) fscanf(indata, "%f %f %f %f %f %f %f %f", &x, &y, &z, &handRoll, &handPitch, &handYaw, &armYaw, &armPitch);
     else fscanf(indata, "%f %f %f %f %f", &x, &y, &z, &armYaw, &armPitch);

     armYaw = armYaw*360.0;
     armPitch = armPitch*360.0;

     // get upper arm vector
     vec = mav_vectorMult(MAV_Y_VECTOR, mav_matrixRPYSet(MAV_ID_MATRIX, 0, armPitch, armYaw));
     vec = mav_vectorScalar(vec, -upperArmLength/mav_vectorMag(vec));
     // shift so relative to back
     vec.x -= shoulderWidth;

     // calculate pitch as before (from -y-axis)
     armPitch = -MAV_RAD2DEG(acos( mav_vectorDotProduct(mav_vectorScalar(MAV_Y_VECTOR,-1),mav_vectorNormalise(vec)) ));
     armPitch += 360;

     // calculate yaw from x-axis (rather than z-axis)
     v = mav_vectorSet(vec.x, 0, vec.z); // reflect onto the xz plane
     armYaw = MAV_RAD2DEG(acos( mav_vectorDotProduct(MAV_X_VECTOR,mav_vectorNormalise(v)) )); // get yaw of arm in xz plane relative to x-vector
     normal = mav_vectorCrossProduct(MAV_X_VECTOR,mav_vectorNormalise(v)); // get normal to x-vector
     cosAngle = mav_vectorDotProduct(mav_vectorNormalise(normal),MAV_Y_VECTOR);
     if ( cosAngle < -0.999999)  // if normal at 180 degrees to y-vector, then negative yaw
       armYaw = -armYaw + 360;

     armYaw = armYaw/360.0;
     armPitch = armPitch/360.0;

     // write new values to file
     if (type==1) fprintf(outdata, "%f %f %f %f %f %f %f %f\n", x, y, z, handRoll, handPitch, handYaw, armYaw, armPitch);
     else fprintf(outdata, "%f %f %f %f %f\n", x, y, z, armYaw, armPitch);
     }

   printf("done\n");
   fclose(indata);
   }
 else
   {
   printf("failed to open '%s'\n", inFileName);
   fclose(outdata);
   return 0;
   }

 fclose(outdata);
 return 1;
 }

/*************************************************************/
/* calculate error (SSE, real, and average) between patterns */
/*************************************************************/
// BUG WARNING: only outputs integer errors (in float format) hence this function is pretty useless at the moment
int calcError(char *inFileName, int type)
 {
 FILE *indata, *outdata, *outdata1, *outdata2, *outdata3;
 char str[200];
 float temp, armPitch, armYaw, predictedArmPitch, predictedArmYaw;
 float error, diffyaw, diffpitch, totalSSE, totalRealYaw, totalRealPitch;
 int count;

 if (type==0 || type==1)
   {
   printf("not calculating error:- must load (testing) results data\n");
   return 0;
   }

 // try to create data file
 strcpy(str, inFileName);
 strcat(str, ".sse");
 if ((outdata = fopen(str,"w"))!=NULL)
   printf("writing SSE errors to '%s'\n", str);
 else
   {
   printf("failed to open '%s' for writing\n", str);
   return 0;
   }

 strcpy(str, inFileName);
 strcat(str, ".sse.ave");
 if ((outdata2 = fopen(str,"w"))!=NULL)
   printf("writing average SSE error to '%s'\n", str);
 else
   {
   printf("failed to open '%s' for writing\n", str);
   fclose(outdata);
   return 0;
   }

 strcpy(str, inFileName);
 strcat(str, ".real");
 if ((outdata1 = fopen(str,"w"))!=NULL)
   printf("writing real errors to '%s'\n", str);
 else
   {
   printf("failed to open '%s' for writing\n", str);
   fclose(outdata);
   fclose(outdata2);
   return 0;
   }

 strcpy(str, inFileName);
 strcat(str, ".real.ave");
 if ((outdata3 = fopen(str,"w"))!=NULL)
   printf("writing average real error to '%s'\n", str);
 else
   {
   printf("failed to open '%s' for writing\n", str);
   fclose(outdata);
   fclose(outdata2);
   fclose(outdata1);
   return 0;
   }

 // calculate errors
 totalSSE = 0.0;
 totalRealYaw = 0.0;
 totalRealPitch = 0.0;
 if ((indata = fopen(inFileName,"r"))!=NULL)
   {
   printf("calculating errors from '%s'...", inFileName);
   count = 0;
   while (!feof(indata))
     {
     count++;
     if (type==2)
       fscanf(indata, "%f %f %f %f", &armYaw, &armPitch, &predictedArmYaw, &predictedArmPitch); // read pattern
     else
       fscanf(indata, "%f %f %f %f %f", &temp, &armYaw, &armPitch, &predictedArmYaw, &predictedArmPitch); // read pattern
     // convert to degrees from radians
     armYaw = MAV_RAD2DEG(armYaw);
     armPitch = MAV_RAD2DEG(armPitch);
     predictedArmYaw = MAV_RAD2DEG(predictedArmYaw);
     predictedArmPitch = MAV_RAD2DEG(predictedArmPitch);
     // calculate differences
     diffyaw = abs(armYaw - predictedArmYaw);
     diffpitch = abs(armPitch - predictedArmPitch);
     totalRealYaw += diffyaw;
     totalRealPitch += diffpitch;
     fprintf(outdata1, "%d %f %f\n", count, diffyaw, diffpitch); // write real yaw and pitch error
     // calculate SSE
     error = ((diffpitch*diffpitch)+(diffyaw*diffyaw)) / 2.0 ;// calculate SSE error
     totalSSE += error;
     fprintf(outdata, "%d %f\n", count, error); // write SSE error
     }

   // write average errors
   fprintf(outdata2, "%f\n", totalSSE/((float) count)); // write ave. SSE
   fprintf(outdata3, "%f %f %f\n", totalRealYaw/((float) count), totalRealPitch/((float) count), ((totalRealYaw/((float) count))+(totalRealPitch/((float) count)))/2.0); // write ave. real yaw, pitch, and both

   printf("done\n");
   fclose(indata);
   }
 else
   {
   printf("failed to open '%s'\n", inFileName);
   fclose(outdata);
   fclose(outdata1);
   fclose(outdata2);
   fclose(outdata3);
   return 0;
   }

 fclose(outdata);
 fclose(outdata1);
 fclose(outdata2);
 fclose(outdata3);
 return 1;
 }

/**************************************************************************************************/

/*****************************************************************/
/* change from arm rotation to elbow position (relative to back) */
/*****************************************************************/

MAV_vector getElbowPoint(float yaw, float pitch, float shoulderWidth, float upperArmLength, float lowerArmLength)
 {
 MAV_vector elbow;
 float cosAngle, angle, dist;
 float a, b, c;

 // convert angles back to range 0->360 (from 0 ->1)
 yaw = yaw*360.0;
 pitch = pitch*360.0;

 if (verbose) printf("\n----- in getElbowPoint -----\n");
 // get target direction of elbow
 elbow = mav_vectorRotate(mav_vectorScalar(MAV_Y_VECTOR,-1), mav_vectorScalar(MAV_Z_VECTOR,-1), MAV_DEG2RAD(pitch));
 if (verbose) printf("getElbowPoint: rotating -Y by %f about -Z to get (%f %f %f),\n", pitch, elbow.x, elbow.y, elbow.z);
 elbow = mav_vectorRotate(elbow, MAV_Y_VECTOR, MAV_DEG2RAD(yaw));
 if (verbose) printf("               then by %f about Y to get (%f %f %f)\n", yaw, elbow.x, elbow.y, elbow.z);

 // extend to get target elbow point (relative to back)
 cosAngle = mav_vectorDotProduct(elbow, mav_vectorScalar(MAV_X_VECTOR,-1));
 angle = acos(cosAngle);
 if (verbose) printf("getElbowPoint: angle = %f\n", MAV_RAD2DEG(angle));

 a = (upperArmLength*upperArmLength);
 b = (shoulderWidth*sin(angle));
 c = (shoulderWidth*cosAngle);
 dist = sqrt(a-(b*b))+c;
// dist = sqrt((upperArmLength*upperArmLength)-((shoulderWidth*sin(angle))*(shoulderWidth*sin(angle))))+(shoulderWidth*cosAngle);
 if (verbose) printf("getElbowPoint: a = %f, b = %f, c = %f\n", a,b,c);
 if (verbose) printf("getElbowPoint: dist = %f\n", dist);
 elbow = mav_vectorScalar(elbow, dist);
 if (verbose) printf("getElbowPoint: extending by %f to get (%f %f %f)\n", dist, elbow.x, elbow.y, elbow.z);
 if (verbose) printf("----- out getElbowPoint -----\n\n");

 return elbow;
 }

MAV_vector IKAlgorithm(MAV_vector prediction, MAV_vector hand, float shoulderWidth, float upperArmLength, float lowerArmLength)
 {
 MAV_vector elbow, normal;
 float angle, cosAngle, correctedAngle, dist, maxMag;

 float totalArmLength = upperArmLength + lowerArmLength;
 elbow = prediction;

 if (verbose) printf("\n----- in IKAlgorithm -----\n");

 // un-normalise hand
 hand.x = (hand.x-0.5)*2.0; // map x onto range -1 -> 1 (from range 0 -> 1)
 hand.y = (hand.y-0.5)*2.0;
 hand.z = (hand.z-0.5)*2.0;
 maxMag = shoulderWidth + upperArmLength + lowerArmLength; // calculate maximum possible hand vector magnitude
 if (verbose) printf("max hand vector magnitude = %f\n", maxMag);
 hand = mav_vectorScalar(hand, maxMag);

 // normalise hand vector
 /*vec = mav_vectorSet(x, y, z);
 tmp = mav_vectorMag(vec);
 vec = mav_vectorScalar(vec, (tmp/maxMag)/tmp);
 vec.x = (vec.x/2.0)+0.5;
 vec.y = (vec.y/2.0)+0.5;
 vec.z = (vec.z/2.0)+0.5;*/

 if (verbose) printf("IKAlgorithm: shifting elbow (%f %f %f) to ", elbow.x, elbow.y, elbow.z);
 // shift predicted elbow and hand position relative to shoulder (rather than back)
 elbow.x += shoulderWidth;
 if (verbose) printf("(%f %f %f)\n", elbow.x, elbow.y, elbow.z);
 if (verbose) printf("IKAlgorithm: shifting hand (%f %f %f) to ", hand.x, hand.y, hand.z);
 hand.x += shoulderWidth;
 if (verbose) printf("(%f %f %f)\n", hand.x, hand.y, hand.z);

 // perform IK correction...
 dist = mav_vectorMag(hand); // get distance of hand from shoulder
 if (verbose) printf("(hand) dist = %f, totalArmLength = %f\n", dist, totalArmLength);
 if ( dist>totalArmLength ) // if distance from shoulder to hand is greater than total arm length
   {
   hand = mav_vectorScalar(hand, totalArmLength/dist); // scale shoulder-to-hand vector to total arm length
   if (verbose) printf("IKAlgorithm: (hand) dist warning, scaling hand to (%f %f %f)\n", hand.x, hand.y, hand.z);
   if (verbose) printf("IKAlgorithm: hand mag = %f\n", mav_vectorMag(hand));
   dist = totalArmLength; // set distance as total arm length
   }
 cosAngle = ((upperArmLength*upperArmLength)+(dist*dist)-(lowerArmLength*lowerArmLength)) / (2.0*upperArmLength*dist); // calculate correct angle
 if (verbose) printf("IKAlgorithm: cosAngle = %f\n", cosAngle);
 correctedAngle = acos(cosAngle);
 if (verbose) printf("IKAlgorithm: correctedAngle = %f\n", MAV_RAD2DEG(correctedAngle));
 // calculate normal and angle between predicted upper arm vector (predicted elbow) and shoulder-to-hand vector
 normal = mav_vectorCrossProduct(mav_vectorNormalise(elbow), mav_vectorNormalise(hand));
 cosAngle = mav_vectorDotProduct(mav_vectorNormalise(elbow), mav_vectorNormalise(hand));
 angle = acos(cosAngle);
 if (verbose) printf("IKAlgorithm: angle = %f\n", MAV_RAD2DEG(angle));
 // correct predicted elbow position
 if (verbose) printf("IKAlgorithm: correcting (%f %f %f) by %f about (%f %f %f) ", elbow.x, elbow.y, elbow.z, MAV_RAD2DEG(angle-correctedAngle), normal.x, normal.y,  normal.z);
 elbow = mav_vectorRotate(elbow, mav_vectorNormalise(normal), angle-correctedAngle); // rotate upper arm vector about normal.
// printf("angular correction = %f\n", MAV_RAD2DEG(angle-correctedAngle));
 if (verbose) printf("to get (%f %f %f)\n",elbow.x, elbow.y, elbow.z);

 // shift back so relative to back
 elbow.x -= shoulderWidth;

 if (verbose) printf("----- out IKAlgorithm -----\n\n");
 return elbow;
 }

// BUG WARNING: Will read last line of input file even if blank and hence calculate a huge error.
//              So user must ensure output file is same length as input files (testing data and network output files).
int angularToPositional(char *netoutFilename, int patternNumbers, char *testingFilename,
                        float shoulderWidth, float upperArmLength, float lowerArmLength, int ikAlgorithm)
 {
 char str[200];
 MAV_vector targetElbow, predictedElbow, v, hand, normal;
 FILE *netout_indata, *testing_indata, *outdata, *outdata1;
 float targetYaw, targetPitch, predictedYaw, predictedPitch, x, y, z, error, totalError, tmp;
 int count;

 // try to create data file to hold new positional data
 strcpy(str, netoutFilename);
 if (ikAlgorithm==1)
   strcat(str, ".pos.ik");
 else
   strcat(str, ".pos");
 if ((outdata = fopen(str,"w"))!=NULL)
   printf("writing elbow positional data to '%s'\n", str);
 else
   {
   printf("failed to open '%s' for writing\n", str);
   return 0;
   }

 strcpy(str, netoutFilename);
 if (ikAlgorithm==1)
   strcat(str, ".pos.ik.ave");
 else
   strcat(str, ".pos.ave");
 if ((outdata1 = fopen(str,"w"))!=NULL)
   printf("writing average error to '%s'\n", str);
 else
   {
   printf("failed to open '%s' for writing\n", str);
   fclose(outdata);
   return 0;
   }

 // start parsing files
 count = 0;
 totalError = 0.0;
 if ((testing_indata = fopen(testingFilename,"r"))!=NULL && (netout_indata = fopen(netoutFilename,"r"))!=NULL)
   {
   printf("processing...\n");
   while (!feof(testing_indata) && !feof(netout_indata))
     {
     count += 1;

     // read netout data to get target and predicted yaw and pitch
     if (patternNumbers==1)
       fscanf(netout_indata, "%f %f %f %f %f", &tmp, &targetYaw, &targetPitch, &predictedYaw, &predictedPitch);
     else
       fscanf(netout_indata, "%f %f %f %f", &targetYaw, &targetPitch, &predictedYaw, &predictedPitch);

     // get target and predicted elbow point
     targetElbow = getElbowPoint(targetYaw, targetPitch, shoulderWidth, upperArmLength, lowerArmLength);
     if (verbose) mav_vectorPrint("targetElbow = ", targetElbow);
     predictedElbow = getElbowPoint(predictedYaw, predictedPitch, shoulderWidth, upperArmLength, lowerArmLength);
     if (verbose) mav_vectorPrint("predictedElbow = ", predictedElbow);

     // perform IK algorithm if desired
     if (ikAlgorithm==1)
       {
       // read testing data to get hand position
       fscanf(testing_indata, "%f %f %f %f %f %f %f %f", &x, &y, &z, &tmp, &tmp, &tmp, &tmp, &tmp);
       hand = mav_vectorSet(x, y, z);

       // perform IK correction algorithm on target and predicted elbow position
//       mav_vectorPrint("targetElbow = ",targetElbow);
       targetElbow = IKAlgorithm(targetElbow, hand, shoulderWidth, upperArmLength, lowerArmLength);
       if (verbose) mav_vectorPrint("corrected targetElbow = ",targetElbow);
//       mav_vectorPrint("predictedElbow = ",predictedElbow);
       predictedElbow = IKAlgorithm(predictedElbow, hand, shoulderWidth, upperArmLength, lowerArmLength);
       if (verbose) mav_vectorPrint("corrected predictedElbow = ",predictedElbow);
       }

     // calculate error (the distance between target and predicted elbow point)
     v = mav_vectorSub(predictedElbow, targetElbow);
     error = mav_vectorMag(v);
     totalError += error;
     if (verbose) printf("error = %f\n", error);

     // write pattern count, target and predicted elbow positions, and error to file
     if (patternNumbers==1) fprintf(outdata, "%d %f %f %f %f %f %f %f\n", count, targetElbow.x, targetElbow.y, targetElbow.z, predictedElbow.x, predictedElbow.y, predictedElbow.z, error);
     else fprintf(outdata, "%f %f %f %f %f %f %f\n", targetElbow.x, targetElbow.y, targetElbow.z, predictedElbow.x, predictedElbow.y, predictedElbow.z, error);
     }

   // write average error
   fprintf(outdata1, "%f\n", totalError/((float) count));

   printf("done\n");
   fclose(testing_indata);
   fclose(netout_indata);
   }
 else
   {
   printf("failed to open either '%s' or '%s'\n", testingFilename, netoutFilename);
   fclose(outdata);
   fclose(outdata1);
   return 0;
   }

 fclose(outdata);
 fclose(outdata1);
 return 1;
 }

/***** temporary function *************************************************************************/

/*************************************/
/* correct data to be MSE (from SSE) */
/*************************************/
/*int tmpCorrect(char *inFileName, int type, char *filename)
 {
 FILE *indata, *outdata;
 char str[200];
 float num, error;
 int tmp, count, length;

 // try to create data file
 strcpy(str, inFileName);
 strcat(str, ".mse");
 if ((outdata = fopen(str,"w"))!=NULL)
   printf("writing corrected error data to '%s'\n", str);
 else
   {
   printf("failed to open '%s' for writing\n", str);
   return 0;
   }

 length = numLines(inFileName);

 num = (float) numLines(filename);
 printf("correcting for %d training patterns...\n", (int)num);
// num = num/sqrt(num);

 if ((indata = fopen(inFileName,"r"))!=NULL)
   {
   printf("correcting %s...", inFileName);
   count = 0;
   while (!feof(indata))
     {
     // read pattern
     fscanf(indata, "%d %f", &tmp, &error);

     if (count++ < length) // this prevents an extra line being created if a carriage return is at end of inFileName
       {
       error = (error*num)*(error*num);
       error = error/num;
       // write new values to file
       fprintf(outdata, "%d %f\n", tmp, error);
       }

     }

   printf("done\n");
   fclose(indata);
   }
 else
   {
   printf("failed to open '%s'\n", inFileName);
   fclose(outdata);
   return 0;
   }

 fclose(outdata);
 return 1;
 }*/
   
/**************************************************************************************************/

int main (char argc, char *argv[])
 {
 char command, choice;
 char inFileName[200]; // filename of data file
 char str[200];
 FILE *datain;

 int type = 0; // type of data in file, set by loadFile function
               // 0 - training data, no hand RPY (epoch hand.x hand.y hand.z armYaw armPitch)
               // 1 - training data, with hand RPY (epoch hand.x hand.y hand.z handRoll handPitch handYaw armYaw armPitch)
               // 2 - (testing) results data, no pattern number (targetYaw targetPitch actualYaw actualPitch)
               // 3 - (testing) results data, with pattern number (patternNumber targetYaw targetPitch actualYaw actualPitch)
               // -1 - neither, so not loaded (but this is never tested to determine whether a file is loaded, the variable fileLoaded is)
 int tmp, tmp1;
 int fileLoaded = 0; // indicates whether a file had been loaded

 float shoulderWidth, upperArmLength, lowerArmLength, val;
 int count;

 while (command != 'q')
   {
   printf("please enter command (h for help): ");
   scanf("%s", &command);

   switch (command)
     {
     case 'q': // quit
       exit(0);
       break;
     case 'h': // display help
       printf("\n--- help ---\n\n");
       printf(" l - load data file\n");
       printf(" v - view data\n");
       printf(" j - join data\n");
       printf(" s - split data\n");
       printf(" a - remove hand angular data (RPY)\n");
       printf(" c - clean data (remove duplicates)\n");
       printf(" n - normalise data\n");
       printf(" r - randomise data\n");
       printf(" b - get arm rotation relative to back (and yaw relative to x-axis) rather\n");
       printf("     than relative to shoulder (and yaw relative to z-axis)\n");
       printf(" e - calculate errors (SSE and real) from testing results data\n");
       printf(" p - translate upper arm yaw and pitch to elbow position (relative to back)\n");
       /*printf(" m - correct error to be MSE rather than incorrect SSE (temporary function)\n");*/
       printf(" h - help\n");
       printf(" q - quit\n");
       printf("\n------------\n\n");
       break;
     case 'l': // load
       if (loadFile(inFileName, &type))
         fileLoaded = 1;
       else
         fileLoaded = 0;
       break;
     case 'n': // normalise
       if (fileLoaded==0) // check that a file is loaded
         {
         printf("press 'l' to load data file\n");
         break;
         }

       // read user's data file to get body proportions
       printf("please enter user file name: ");
       scanf("%s", str);
       if ((datain = fopen(str,"r"))!=NULL)
         {
         printf("reading user's dimensions from '%s'\n", str);
         //if (!feof(datain)){ fscanf(datain, "%f", &val); } // if neck length in user file
         if (!feof(datain)){ fscanf(datain, "%f", &shoulderWidth); }
         if (!feof(datain)){ fscanf(datain, "%f", &upperArmLength); }
         if (!feof(datain)){ fscanf(datain, "%f", &lowerArmLength); }
         printf("shoulder width = %f\n", shoulderWidth);
         printf("upper arm length = %f\n", upperArmLength);
         printf("lower arm length = %f\n", lowerArmLength);
         fclose(datain); // close file
         }
       else
         {
         printf("failed to open '%s' for reading\n", str);
         break;
         }

       // normalise
       if (normalise(inFileName, type, shoulderWidth, upperArmLength, lowerArmLength))
         {
         // update current filename to be normalised file
         strcat(inFileName, ".norm");
         printf("current working file is %s\n", inFileName);
         }

       break;
     case 's': // split
       if (fileLoaded==0) // check that a file is loaded
         {
         printf("press 'l' to load data file\n");
         break;
         }

       // prompt for percentage split
       printf("please enter percentage split (1-100): ");
       scanf("%f", &val);

       // split
       if (split(inFileName, val))
         {
         // update current filename
         strcat(inFileName, ".split1");
         printf("current working file is %s\n", inFileName);
         }

       break;
     case 'j': // join
       if (fileLoaded==0) // check that a file is loaded
         {
         printf("press 'l' to load data file\n");
         break;
         }

       // try to load file to join with
       printf("please enter details of file to join with '%s'...\n", inFileName);
       if (loadFile(str, &tmp))
         {
         if (join(inFileName, str)) // join
           {
           // update current filename
           strcat(inFileName, ".join");
           printf("current working file is %s\n", inFileName);
           }
         }

       break;
     case 'r': // randomise
       if (fileLoaded==0) // check that a file is loaded
         {
         printf("press 'l' to load data file\n");
         break;
         }

       // randomise
       if (randomise(inFileName))
         {
         // update current filename
         strcat(inFileName, ".rand");
         printf("current working file is %s\n", inFileName);
         }

       break;
     case 'a': // remove hand angular info (RPY)
       if (fileLoaded==0) // check that a file is loaded
         {
         printf("press 'l' to load data file\n");
         break;
         }

       // remove angular info
       if (removeHandRPY(inFileName, type))
         {
         // update current filename
         strcat(inFileName, ".noang");
         printf("current working file is %s\n", inFileName);
         }

       break;
     case 'c': // clean data
       if (fileLoaded==0) // check that a file is loaded
         {
         printf("press 'l' to load data file\n");
         break;
         }

       // prompt sensitivity (difference between hand data)
       //printf("please enter sensitivity: ");
       //scanf("%f", &val);

       // clean patterns
       if (clean(inFileName, type))
         {
         // update current filename to be cleaned file
         strcat(inFileName, ".clean");
         printf("current working file is %s\n", inFileName);
         }

       break;
     case 'b': // get arm rotation relative to back, etc...
       if (fileLoaded==0) // check that a file is loaded
         {
         printf("press 'l' to load data file\n");
         break;
         }

        printf("has data been normalised? (y or n): ");
        scanf("%s", &choice);
        if (choice!='y')
          {
          printf("not changing arm rotation:- data must be normalised first\n");
          break;
          }

       // read user's data file to get body proportions
       printf("please enter user file name: ");
       scanf("%s", str);
       if ((datain = fopen(str,"r"))!=NULL)
         {
         printf("reading user's dimensions from '%s'\n", str);
         //if (!feof(datain)){ fscanf(datain, "%f", &val); } // if neck length in user file
         if (!feof(datain)){ fscanf(datain, "%f", &shoulderWidth); }
         if (!feof(datain)){ fscanf(datain, "%f", &upperArmLength); }
         if (!feof(datain)){ fscanf(datain, "%f", &lowerArmLength); }
         printf("shoulder width = %f\n", shoulderWidth);
         printf("upper arm length = %f\n", upperArmLength);
         printf("lower arm length = %f\n", lowerArmLength);
         fclose(datain); // close file
         }
       else
         {
         printf("failed to open '%s' for reading\n", str);
         break;
         }

       // get arm rotation relative to back (and yaw relative to x-axis)
       if (relBack(inFileName, type, shoulderWidth, upperArmLength))
         {
         // update current filename to be processed file
         strcat(inFileName, ".back");
         printf("current working file is %s\n", inFileName);
         }

       break;
     case 'e': // calculate SSE and real error
       if (fileLoaded==0) // check that a file is loaded
         {
         printf("press 'l' to load data file\n");
         break;
         }

       // calculate error
       if (calcError(inFileName, type))
         {
         printf("current working file is still %s\n", inFileName);
         }

       break;
     case 'p': // translate from upper arm yaw and pitch, to elbow position

       printf("please note that this function assumes all data to be relative to the back and the testing data to contain hand RPY\n");

       // read user's data file to get body proportions
       printf("please enter user file name: ");
       scanf("%s", str);
       if ((datain = fopen(str,"r"))!=NULL)
         {
         printf("reading user's dimensions from '%s'\n", str);
         //if (!feof(datain)){ fscanf(datain, "%f", &val); } // if neck length in user file
         if (!feof(datain)){ fscanf(datain, "%f", &shoulderWidth); }
         if (!feof(datain)){ fscanf(datain, "%f", &upperArmLength); }
         if (!feof(datain)){ fscanf(datain, "%f", &lowerArmLength); }
         printf("shoulder width = %f\n", shoulderWidth);
         printf("upper arm length = %f\n", upperArmLength);
         printf("lower arm length = %f\n", lowerArmLength);
         fclose(datain); // close file
         }
       else
         {
         printf("failed to open '%s' for reading\n", str);
         break;
         }

       // try to load testing data file
       printf("please enter name of testing data file: ");
       scanf("%s", str);
       if (fopen(str,"r")!=NULL)
         {
         printf("perform IK algorithm? (y or n): ");
         scanf("%s", &choice);
         if (choice=='y')
           {
           tmp1 = 1;
           }
         else
           {
           tmp1 = 0;
           }
         }
       else
         {
         printf("failed to open '%s' for reading\n", str);
         break;
         }

       // try to load network output (targets outputs) file
       printf("please enter name of network output file: ");
       scanf("%s", inFileName);
       if (fopen(inFileName,"r")!=NULL)
         {
         printf("does file specify pattern numbers? (y or n): ");
         scanf("%s", &choice);
         if (choice=='y')
           {
           tmp = 1;
           }
         else
           {
           tmp = 0;
           }
         // perform translation
         angularToPositional(inFileName, tmp, str, shoulderWidth, upperArmLength, lowerArmLength, tmp1);
         }
       else
         {
         printf("failed to open '%s' for reading\n", inFileName);
         break;
         }

       break;
     /*case 'm': // correct error to be MSE
       if (fileLoaded==0) // check that a file is loaded
         {
         printf("press 'l' to load data file\n");
         break;
         }

       // try to load file containing the data that the network was trained on (need to know the number of patterns)
       printf("please enter file name of data that trained the NN...\n", str);
       if (loadFile(str, &tmp))
         {
         if (tmp!=0 && tmp!=1)
           {
           printf("not correcting errors:- must specify training data\n");
           break;
           }
         // change error to RMS error
         if (tmpCorrect(inFileName, type, str))
           {
           // update current filename
           strcat(inFileName, ".mse");
           printf("current working file is %s\n", inFileName);
           }
         }

       break;*/
     case 'v':
       if (fileLoaded==0) // check that a file is loaded
         {
         printf("press 'l' to load data file\n");
         break;
         }

       if ((datain = fopen(inFileName,"r"))!=NULL)
         {
         count = 0;
         printf("-- begin --\n");
         while (fgets(str,200,datain) != NULL) // read line
           {
           count++;
           printf("pattern %d = %s", count, str);
           }
         printf("-- end --\n");
         }
       else
         {
         printf("failed to open '%s' for reading\n", str);
         break;
         }

       break;
     }
   }// end while

 return 0;
 }

/**************************************************************************************************/

